@@ -54,9 +54,16 @@ def build_cache(cli):
    bu.cache.summary_print()
 
 def delete(cli):
-   img = ch.Image(ch.Image_Ref(cli.image_ref))
-   img.unpack_delete()
-   bu.cache.worktrees_prune()
+   delete_ct = 0
+   for img in ch.Image.glob(cli.image_ref):
+      img.unpack_delete()
+      delete_ct += 1
+   for img in ch.Image.glob(cli.image_ref + "_stage[0-9]*"):
+      img.unpack_delete()
+      delete_ct += 1
+   if (delete_ct == 0):
+      ch.FATAL("no image matching glob, can’t delete: %s" % cli.image_ref)
+   bu.cache.worktrees_fix()
 
 def gestalt_bucache(cli):
    bu.have_deps()
@@ -93,8 +100,15 @@ def list_(cli):
          ch.FATAL("does not exist: %s" % ch.storage.root)
       if (not ch.storage.valid_p):
          ch.FATAL("not a storage directory: %s" % ch.storage.root)
-      for img in sorted(ch.listdir(imgdir)):
-         print(ch.Image_Ref(img))
+      images = sorted(imgdir.listdir())
+      if (len(images) >= 1):
+         img_width = max(len(ref) for ref in images)
+         for ref in images:
+            img = ch.Image(ch.Image_Ref(ch.Path(ref).parts[-1]))
+            if cli.long:
+               print("%-*s | %s" % (img_width, img, img.last_modified.ctime()))
+            else:
+               print(img)
    else:
       # list specified image
       img = ch.Image(ch.Image_Ref(cli.image_ref))
@@ -104,7 +118,8 @@ def list_(cli):
          stored = "no"
       else:
          img.metadata_load()
-         stored = "yes (%s)" % img.metadata["arch"]
+         stored = "yes (%s), modified: %s" % (img.metadata["arch"],
+                                              img.last_modified.ctime())
       print("in local storage:    %s" % stored)
       # in cache?
       (sid, commit) = bu.cache.find_image(img)
@@ -126,21 +141,43 @@ def list_(cli):
          pullet.fatman_load()
          remote = "yes"
          arch_aware = "yes"
-         arch_avail = " ".join(sorted(pullet.architectures.keys()))
+         arch_keys = sorted(pullet.architectures.keys())
+         try:
+            fmt_space = len(max(arch_keys,key=len))
+            arch_avail = []
+            for key in arch_keys:
+               arch_avail.append("%-*s  %s" % (fmt_space, key,
+                                               pullet.digests[key][:11]))
+         except ValueError:
+            # handles case where arch_keys is empty, e.g.
+            # mcr.microsoft.com/windows:20H2.
+            arch_avail = [None]
       except ch.Image_Unavailable_Error:
          remote = "no (or you are not authorized)"
          arch_aware = "n/a"
-         arch_avail = "n/a"
+         arch_avail = ["n/a"]
       except ch.No_Fatman_Error:
          remote = "yes"
          arch_aware = "no"
-         arch_avail = "unknown"
+         arch_avail = ["unknown"]
       pullet.done()
       print("available remotely:  %s" % remote)
       print("remote arch-aware:   %s" % arch_aware)
       print("host architecture:   %s" % ch.arch_host)
-      print("archs available:     %s" % arch_avail)
+      print("archs available:     %s" % arch_avail[0])
+      for arch in arch_avail[1:]:
+         print((" " * 21) + arch)
 
 def reset(cli):
    ch.storage.reset()
 
+def undelete(cli):
+   if (cli.bucache != ch.Build_Mode.ENABLED):
+      ch.FATAL("only available when cache is enabled")
+   img = ch.Image(ch.Image_Ref(cli.image_ref))
+   if (img.unpack_exist_p):
+      ch.FATAL("image exists; will not overwrite")
+   (_, git_hash) = bu.cache.find_image(img)
+   if (git_hash is None):
+      ch.FATAL("image not in cache")
+   bu.cache.checkout(img, git_hash, None)
